package org.openarchitectureware.adapter.rsdp.workflow;

import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.uml2.common.util.UML2Util;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.Profile;
import org.openarchitectureware.workflow.WorkflowContext;
import org.openarchitectureware.workflow.issues.Issues;
import org.openarchitectureware.workflow.lib.AbstractWorkflowComponent;
import org.openarchitectureware.workflow.monitor.ProgressMonitor;

/**
 * 
 * @author Benedikt Niehues
 * 
 */
public class EMXWriter extends AbstractWorkflowComponent {

	private static final String COMPONENT_NAME = "EMX Writer";

	private String outPath;

	private String destinationFolder = "/";

	private String inputSlot = WorkflowContext.DEFAULT_SLOT;

	/**
	 * Returns the name of the destination folder.
	 * 
	 * @return name of destination folder
	 */
	public String getDestinationFolder() {
		return destinationFolder;
	}

	/**
	 * Sets the name of the destination folder
	 * 
	 * @param destinationFolder
	 *            name of folder
	 */
	public void setDestinationFolder(final String destinationFolder) {
		this.destinationFolder = destinationFolder;
		if (!this.destinationFolder.equals("")) {
			if (!this.destinationFolder.endsWith("/")) {
				this.destinationFolder += "/";
			}
		}
		else {
			this.destinationFolder = "/";
		}
	}

	/**
	 * Sets the name of the input slot.
	 * 
	 * @param slot
	 *            name of slot
	 */
	public void setInputSlot(final String slot) {
		this.inputSlot = slot;
	}

	/**
	 * Sets the output path.
	 * 
	 * @param path
	 *            the output path
	 */
	public void setOutPath(final String path) {
		this.outPath = path;
		if (!this.outPath.endsWith("/")) {
			this.outPath += "/";
		}
		if (this.outPath.startsWith("/")) {
			this.outPath = this.outPath.substring(1);
		}
	}

	/**
	 * Returns the name of the input slot.
	 * 
	 * @return name of input slot
	 */
	public String getInputSlot() {
		return inputSlot;
	}

	/**
	 * Returns the output path.
	 * 
	 * @return the output path
	 */
	public String getOutPath() {
		return outPath;
	}

	public void checkConfiguration(final Issues issues) {

	}

	private static Log log = LogFactory.getLog(EMXWriter.class);

	@SuppressWarnings("unchecked")
	public void invoke(final WorkflowContext ctx, final ProgressMonitor monitor, final Issues issues) {
		log.info("Running " + getComponentName());
		final Object slotContent = ctx.get(inputSlot);
		if (slotContent == null) {
			issues.addError(this, "slot '" + inputSlot + "' is empty.");
			return;
		}
		if (!(slotContent instanceof Collection<?> || slotContent instanceof org.eclipse.uml2.uml.Package)) {
			issues
					.addError(this, "slot '" + inputSlot + "' does not contain a list, a model or a profile",
							slotContent);
			return;
		}
		ResourceSetImpl rsImpl = new ResourceSetImpl();
		// this is for Rational Software Modeler
		Resource.Factory factory = (Resource.Factory) Resource.Factory.Registry.INSTANCE.getExtensionToFactoryMap()
				.get("uml2");
		rsImpl.getResourceFactoryRegistry().getExtensionToFactoryMap().put("emx", factory);
		rsImpl.getResourceFactoryRegistry().getExtensionToFactoryMap().put("epx", factory);

		if (slotContent instanceof Collection<?>) {
			Iterator it = ((Collection) slotContent).iterator();
			while (it.hasNext()) {
				Object o = it.next();
				if (o instanceof Package) {
					Package model = (Package) o;
					saveModel(model, rsImpl);
				}
			}
		}
		else if (slotContent instanceof Package) {
			saveModel((Package) slotContent, rsImpl);
		}
	}

	@SuppressWarnings("unchecked")
	private void saveModel(Package model, ResourceSet rsImpl) {
		String fileExtension = ".emx";
		if (model instanceof Profile) {
			fileExtension = ".epx";
		}
		String modelFile = destinationFolder + outPath + model.getName() + fileExtension;
		final URI fileURI = URI.createFileURI(modelFile);
		Resource r = rsImpl.createResource(fileURI);
		r.getContents().add(model);
		for (Iterator allContents = UML2Util.getAllContents(model, true, false); allContents.hasNext();) {
			EObject eObject = (EObject) allContents.next();
			if (eObject instanceof Element) {
				r.getContents().addAll(((Element) eObject).getStereotypeApplications());
			}
		}
		try {
			r.save(null);
		}
		catch (IOException e) {
		}
	}

	/**
	 * @see org.openarchitectureware.workflow.WorkflowComponent#getComponentName()
	 */
	public String getComponentName() {
		return COMPONENT_NAME;
	}
}
