/*
 * <copyright>
 *
 * Copyright (c) 2005-2006 Sven Efftinge (http://www.efftinge.de) and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Sven Efftinge (http://www.efftinge.de) - Initial API and implementation
 *
 * </copyright>
 */
package org.openarchitectureware.core.metamodel.jdt;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.jdt.core.ElementChangedEvent;
import org.eclipse.jdt.core.IElementChangedListener;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.openarchitectureware.expression.TypeSystem;
import org.openarchitectureware.expression.parser.SyntaxConstants;
import org.openarchitectureware.internal.OawLog;
import org.openarchitectureware.type.MetaModel;
import org.openarchitectureware.type.Type;
import org.openarchitectureware.workflow.util.Cache;
import org.openarchitectureware.workflow.util.Pair;

public class JdtMetaModel implements MetaModel, IElementChangedListener {

	protected TypeSystem typesystem;

	private final String name;

	private final JdtTypeStrategy strategy;

	private final IJavaProject project;

	/**
	 * @see org.openarchitectureware.type.MetaModel#getTypeSystem()
	 */
	public TypeSystem getTypeSystem() {
		return typesystem;
	}

	/**
	 * @see org.openarchitectureware.type.MetaModel#setTypeSystem(org.openarchitectureware.expression.TypeSystem)
	 */
	public void setTypeSystem(final TypeSystem typeSystem) {
		typesystem = typeSystem;
	}

	private final static Map<IPath, JdtMetaModel> metaModels = new HashMap<IPath, JdtMetaModel>();

	public boolean changed = false;

	/**
	 * @see org.eclipse.jdt.core.IElementChangedListener#elementChanged(org.eclipse.jdt.core.ElementChangedEvent)
	 */
	public void elementChanged(final ElementChangedEvent event) {
		if (!changed && project.isOnClasspath(event.getDelta().getElement())) {
			changed = true;
		}
	}

	/**
	 * Creates a new metamodel object.
	 * 
	 * @param name
	 *            name of metamodel
	 * @param project
	 *            the java project
	 * @param strategy
	 *            the type strategy
	 * @return the newly created metamodel object
	 */
	public final static JdtMetaModel create(final String name, final IJavaProject project,
			final JdtTypeStrategy strategy) {
		JdtMetaModel mm = JdtMetaModel.metaModels.get(project.getPath());
		if (mm == null || mm.changed) {
			if (mm != null) {
				JavaCore.removeElementChangedListener(mm);
			}
			mm = new JdtMetaModel(name, project, strategy);
			JavaCore.addElementChangedListener(mm);
			JdtMetaModel.metaModels.put(project.getPath(), mm);
		}
		return mm;
	}

	public JdtMetaModel(final String name, final IJavaProject project, final JdtTypeStrategy strategy) {
		this.name = name;
		this.strategy = strategy;
		this.project = project;
	}

	/**
	 * Returns the corresponding type of the metamodel for <code>IType</code>
	 * objects
	 * 
	 * @param type
	 *            the <code>IType</code> object
	 * @return the corresponding metamodel type
	 */
	private Type getTypeForIType(final IType type) {
		if (type == null || !type.exists())
			return null;
		try {
			final ITypeHierarchy hierarchy = type.newSupertypeHierarchy(new NullProgressMonitor());
			if (hierarchy.contains(type.getJavaProject().findType(List.class.getName())))
				return typesystem.getListType(typesystem.getObjectType());
			else if (hierarchy.contains(type.getJavaProject().findType(Set.class.getName())))
				return typesystem.getSetType(typesystem.getObjectType());
			else if (hierarchy.contains(type.getJavaProject().findType(Collection.class.getName())))
				return typesystem.getCollectionType(typesystem.getObjectType());
		}
		catch (final JavaModelException e) {
			OawLog.logError(e);
			return null;
		}
		return new JdtTypeImpl(JdtMetaModel.this, type, getName(type), strategy);
	}

	protected Type getType(final String fqn) throws JavaModelException {
		if (fqn.indexOf(SyntaxConstants.NS_DELIM) == -1)
			// namespace
			return null;
		final String typeName = fqn.replaceAll(SyntaxConstants.NS_DELIM, ".");
		final IType type = findType(project, typeName);
		if (type != null)
			return getTypeForIType(type);
		else
			return null;
	}

	private String getName(final IType class1) {
		return class1.getFullyQualifiedName().replaceAll("\\.", SyntaxConstants.NS_DELIM);
	}

	private final Cache<String, Type> typeNameCache = new Cache<String, Type>() {
		@Override
		protected Type createNew(String typeName) {
			try {
				Type result = getType(typeName);
				return result;
			}
			catch (Exception e) {
				return null;
			}
		}
	};

	/**
	 * @see org.openarchitectureware.type.MetaModel#getTypeForName(java.lang.String)
	 */
	public Type getTypeForName(final String typeName) {
		return typeNameCache.get(typeName);
	}

	/**
	 * @see org.openarchitectureware.type.MetaModel#getType(java.lang.Object)
	 */
	public Type getType(final Object obj) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Returns the corresponding type of the metamodel for a class.
	 * 
	 * @param clazz
	 *            the class
	 * @return the corresponding metamodel type
	 */
	public Type getTypeForClass(final IType clazz) {
		return getTypeSystem().getTypeForName(getName(clazz));
	}

	/**
	 * @see org.openarchitectureware.type.MetaModel#getKnownTypes()
	 */
	@SuppressWarnings("unchecked")
	public Set<? extends Type> getKnownTypes() {
		final Set<Type> result = new HashSet(typeNameCache.getValues());
		result.remove(null);
		return result;
	}

	/**
	 * Returns the name of the metamodel.
	 * 
	 * @return name of metamodel
	 */
	public String getName() {
		return name;
	}

	private final Cache<Pair<String, IType>, String> signCache = new Cache<Pair<String, IType>, String>() {

		@Override
		protected String createNew(Pair<String, IType> p) {
			String signature = p.getFirst();
			IType usingType = p.getSecond();
			String type = JdtSignatureDecoder.decodeTypeSignature(signature, usingType);
			return type;
		}
	};

	/**
	 * Returns the fully qualified name of a type.
	 * 
	 * @param signature
	 *            the type signature
	 * @param usingType
	 *            the type
	 * @return the fully qualified name
	 */
	public String getFullyQualifiedName(final String signature, final IType usingType) {
		return signCache.get(new Pair<String, IType>(signature, usingType));
	}

	/**
	 * Finds the corresponding type for a type name.
	 * 
	 * @param project
	 *            the Java project
	 * @param typeName
	 *            the type name
	 * @return the corresponding type
	 * @throws JavaModelException
	 *             Thrown if an invalid query is executed.
	 */
	public IType findType(final IJavaProject project, final String typeName) throws JavaModelException {
		IType t = project.findType(typeName);
		if (t == null) {
			final String[] projects = project.getRequiredProjectNames();
			for (int i = 0; i < projects.length && t == null; i++) {
				final IJavaProject anotherProject = project.getJavaModel().getJavaProject(projects[i]);
				if (anotherProject != null) {
					t = anotherProject.findType(typeName);
				}
			}
		}
		return t;
	}

	public Set<String> getNamespaces() {
		return new HashSet<String>();
	}

}
