/*******************************************************************************
 * Copyright (c) 2005 - 2007 committers of openArchitectureWare and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     committers of openArchitectureWare - initial API and implementation
 *******************************************************************************/
package org.openarchitectureware.debug.communication;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.ServerSocket;
import java.net.Socket;

import org.openarchitectureware.debug.communication.packets.HandshakePacket;
import org.openarchitectureware.debug.communication.packets.AbstractPacket;

/**
 * This class implements a socket connection model. "Packets" (units of information) can be transfered between two
 * different virtual machines. During the initialisation of a socket connection a reader and a writer thread are
 * established. This class can be used both on the sender and receiver side.
 * 
 * @author Clemens Kadura (zAJKa)
 */
public class Connection {

	private ServerSocket ssocket;

	private Socket socket;

	private DataOutputStream out;

	private DataInputStream in;

	private PacketSender sender;

	private PacketReceiver receiver;

	// -------------------------------------------------------------------------
	// *the* main operation methods

	public AbstractPacket listenForPacket(Class<? extends AbstractPacket> type) throws InterruptedIOException {
		return receiver.getPacket(type, -1);
	}

	public AbstractPacket listenForPacket(Class<? extends AbstractPacket> type, int refId) throws InterruptedIOException {
		return receiver.getPacket(type, refId, -1);
	}

	public int sendPacket(AbstractPacket packet) throws IOException {
		if (sender == null)
			return -1;
		return sender.sendPacket(packet);
	}

	// -------------------------------------------------------------------------
	// socket communication handling

	/**
	 * create a new <code>ServerSocket</code>.
	 * 
	 * @param port the communication port
	 * @throws IOException
	 */
	public void startListeningSocket(int port) throws IOException {
		ssocket = new ServerSocket(port);
	}

	/**
	 * the server listens for a client to connect, creates new reader and writer threads and sends a first
	 * handshake packet to test the communication
	 * 
	 * @param timeout
	 * @throws IOException
	 */
	public void accept(int timeout) throws IOException {
		ssocket.setSoTimeout(timeout);
		// will throw SocketTimeoutException, if nobody connects
		socket = ssocket.accept();
		establishReaderAndWriter();
		if (!writeHandshake())
			throw new IOException("handshake failed");
	}

	/**
	 * the client establishes the connection here, creates new reader and writer threads and waits for a handshake
	 * packet to be received from the server
	 * 
	 * @param port
	 * @throws IOException
	 */
	public void connect(int port) throws IOException {
		socket = new Socket("localhost", port);
		establishReaderAndWriter();
		replyHandshake();
	}

	/**
	 * @return if the socket connection is still active
	 */
	public boolean isConnected() {
		return socket != null && !socket.isClosed();
	}

	/**
	 * close and dispose the socket and the (possibly waiting) sender thread
	 */
	public synchronized void close() {
		if (sender != null) {
			try {
				sender.close();
			} catch (Exception e) {
				// IOExceptions may occur
			} finally {
				sender = null;
			}
		}
			
		if (socket != null) {
			try {
				out.close();
				in.close();
				socket.close();
			} catch (Exception e) {
				// IOExceptions may occur
			} finally {
				out = null;
				in = null;
				socket = null;
			}
		}
		
		if (ssocket!=null) {
			try {
				ssocket.close();
			} catch (Exception e) {
				// IOExceptions may occur
			} finally {
				ssocket = null;
			}
		}
	}

	// -------------------------------------------------------------------------
	// the real sending and receiving of packets to be called from PacketReceiver and PacketSender only

	protected AbstractPacket readPacket() throws IOException {
		String className = in.readUTF();
		AbstractPacket packet = instantiatePacket(className);
		packet.readContent(in);
		// System.out.println(Thread.currentThread().getName() + "-RECEIVED-: " + packet);
		return packet;
	}

	@SuppressWarnings("unchecked")
	private AbstractPacket instantiatePacket(String className) throws IOException {
		Class<? extends AbstractPacket> packetClass = null;
		AbstractPacket packet = null;
		String msg = null;
		try {
			packetClass = (Class<? extends AbstractPacket>) Class.forName(className);
		} catch (ClassNotFoundException e) {
			msg = "Couldn't find " + className + " in the classpath.";
		}

		if (msg == null) {
			Constructor c = packetClass.getConstructors()[0];
			Class[] parmTypes = c.getParameterTypes();
			Object[] initargs = new Object[parmTypes.length];
			for (int i = 0; i < parmTypes.length; i++)
				if (parmTypes[i] == int.class)
					initargs[i] = 0;
			try {
				packet = (AbstractPacket) c.newInstance(initargs);
			} catch (IllegalArgumentException e) {
				msg = "Couldn't instantiate " + c + " : " + e;
			} catch (InstantiationException e) {
				msg = "Couldn't instantiate " + c + " : " + e;
			} catch (IllegalAccessException e) {
				msg = "Couldn't instantiate " + c + " : " + e;
			} catch (InvocationTargetException e) {
				msg = "Problem during instantiation of " + c + " : " + e.getTargetException();
			}
		}

		if (msg != null) {
			System.err.println(msg);
			throw new IOException(msg);
		}
		return packet;
	}

	protected void writePacket(AbstractPacket packet) throws IOException {
		out.writeUTF(packet.getClass().getName());
		packet.writeContent(out);
		// System.out.println(Thread.currentThread().getName() + "-SENT-----: " + packet);
	}

	// ********************************************************* private methods

	private void establishReaderAndWriter() throws IOException {
		out = new DataOutputStream(socket.getOutputStream());
		in = new DataInputStream(socket.getInputStream());

		// start receiver and sender in extra threads
		receiver = PacketReceiver.newPacketReceiver(this);
		sender = PacketSender.newPacketSender(this);

	}

	private boolean writeHandshake() throws IOException {
		AbstractPacket packet = new HandshakePacket();
		sendPacket(packet);
		return listenForPacket(HandshakePacket.class).getClass().equals(HandshakePacket.class);
	}

	private void replyHandshake() throws IOException {
		if (listenForPacket(HandshakePacket.class).getClass().equals(HandshakePacket.class))
			sendPacket(new HandshakePacket());
		else
			throw new IOException("handshake failed");
	}

}
