/*******************************************************************************
 * Copyright (c) 2005 - 2007 committers of openArchitectureWare and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     committers of openArchitectureWare - initial API and implementation
 *******************************************************************************/
package org.openarchitectureware.debug.processing.handlers;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.openarchitectureware.debug.communication.Connection;
import org.openarchitectureware.debug.communication.packets.BreakpointPacket;
import org.openarchitectureware.debug.model.SyntaxElementTO;
import org.openarchitectureware.debug.processing.DebugMonitor;
import org.openarchitectureware.debug.processing.IElementAdapter;
import org.openarchitectureware.debug.processing.IRuntimeHandler;
import org.openarchitectureware.debug.processing.IProcessHandler;

/**
 * This class handles the communication of Breakpoints on the runtime side. It listens in an extra thread for set
 * and removal of breakpoints. The <code>DebugMonitor</code> uses this class to suspend the runtime process at
 * breakpoints.
 * 
 * @author Clemens Kadura (zAJKa)
 */
public class BreakpointRuntimeHandler implements IRuntimeHandler, IProcessHandler, Runnable {

	public static final int SET = 1;

	public static final int REMOVE = 2;

	protected Connection connection;

	protected DebugMonitor monitor;

	private List<Object> breakpoints = new ArrayList<Object>();

	private List<SyntaxElementTO> breakpointTOs = new ArrayList<SyntaxElementTO>();

	private List<SyntaxElementTO> toBeRemovedTOs = new ArrayList<SyntaxElementTO>();

	// -------------------------------------------------------------------------

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.openarchitectureware.debug.processing.IRuntimeHandler#init(org.openarchitectureware.debug.processing.DebugMonitor,
	 *      org.openarchitectureware.debug.communication.Connection)
	 */
	public void init(DebugMonitor monitor, Connection connection) {
		this.monitor = monitor;
		this.connection = connection;
		if (monitor != null) {
			monitor.addProcessHandler(this);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.openarchitectureware.debug.processing.IRuntimeHandler#startListener()
	 */
	public void startListener() {
		Thread thread = new Thread(this, getClass().getSimpleName());
		thread.setDaemon(true);
		thread.start();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		try {
			while (true)
				listenAndDispatchCommand();
		} catch (IOException e) {
		}
	}

	private void listenAndDispatchCommand() throws IOException {
		handle((BreakpointPacket) connection.listenForPacket(BreakpointPacket.class));
	}

	private void handle(BreakpointPacket packet) {
		switch (packet.type) {
		case SET:
			doSet(packet.se, null, 0);
			break;
		case REMOVE:
			doRemove(packet.se, null, 0);
			break;
		default:
			break;
		}
	}

	// -------------------------------------------------------------------------

	private void doSet(SyntaxElementTO se, Object actual, int flag) {
		IElementAdapter adapter = monitor.getAdapter(se);
		if (adapter == null)
			return;
		Object element = adapter.findElement(se, actual, flag);
		// breakpoints may be set before the syntax element structure is instantiated
		// in this case we store the SyntaxElementTO and try it again during shallSuspend(...)
		if (element == null){
			breakpointTOs.add(se);
			for (SyntaxElementTO cand : toBeRemovedTOs)
				if(se.equalsBP(cand)){
					toBeRemovedTOs.remove(cand);
					break;
			}
		}else
			breakpoints.add(element);
	}

	private void doRemove(SyntaxElementTO se, Object actual, int flag) {
		IElementAdapter adapter = monitor.getAdapter(se);
		if (adapter == null)
			return;
		Object element = adapter.findElement(se, actual, flag);
		if (element == null){
			toBeRemovedTOs.add(se);
			for (SyntaxElementTO cand : breakpointTOs)
				if(se.equalsBP(cand)){
					breakpointTOs.remove(cand);
					break;
			}
		}else
			breakpoints.remove(element);
	}

	// -------------------------------------------------------------------------
	// process listener implementation

	public boolean isLastCall() {
		return false;
	}

	/**
	 * returns true if a breakpoint is rgeistered for that element
	 * 
	 * @see org.openarchitectureware.debug.processing.IProcessHandler#shallSuspend(boolean, java.lang.Object, int)
	 */
	public boolean shallSuspend(boolean lastState, Object element, int flag) {
		if (!toBeRemovedTOs.isEmpty()) {
			List<SyntaxElementTO> temp = new ArrayList<SyntaxElementTO>();
			temp.addAll(toBeRemovedTOs);
			toBeRemovedTOs.clear();
			for (SyntaxElementTO se : temp)
				doRemove(se, element, flag);
		}
		if (!breakpointTOs.isEmpty()) {
			List<SyntaxElementTO> temp = new ArrayList<SyntaxElementTO>();
			temp.addAll(breakpointTOs);
			breakpointTOs.clear();
			for (SyntaxElementTO se : temp)
				doSet(se, element, flag);
		}
		return lastState || breakpoints.contains(element);
	}

	/**
	 * no contribution here
	 * 
	 * @see org.openarchitectureware.debug.processing.IProcessHandler#shallHandle(boolean, java.lang.Object, int)
	 */
	public boolean shallHandle(boolean lastState, Object element, int flag) {
		return lastState;
	}

	/**
	 * no contribution here
	 * 
	 * @see org.openarchitectureware.debug.processing.IProcessHandler#shallInterrupt(boolean)
	 */
	public boolean shallInterrupt(boolean lastState) {
		return lastState;
	}

}
