/*
 * <copyright>
 *
 * Copyright (c) 2005-2006 Sven Efftinge (http://www.efftinge.de) and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Sven Efftinge (http://www.efftinge.de) - Initial API and implementation
 *
 * </copyright>
 */
package org.openarchitectureware.xpand2.output;

import java.util.HashMap;
import java.util.Map;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.openarchitectureware.expression.ast.SyntaxElement;
import org.openarchitectureware.workflow.ConfigurationException;
import org.openarchitectureware.workflow.util.Pair;
import org.openarchitectureware.xpand2.XpandExecutionContext;
import org.openarchitectureware.xpand2.ast.TextStatement;

/**
 * *
 * 
 * @author Sven Efftinge (http://www.efftinge.de) *
 */
public class OutputImpl implements Output {

	private boolean automaticHyphenation = false;

	public void setAutomaticHyphens(boolean automaticHyphenation) {
		this.automaticHyphenation = automaticHyphenation;
	}

	protected Stack<FileHandle> fileHandles = new Stack<FileHandle>();

	private final Map<String, Outlet> outlets = new HashMap<String, Outlet>();

	public void addOutlet(final Outlet outlet) {
		if (outlets.containsKey(outlet.getName())) {
			if (outlet.getName() == null)
				throw new ConfigurationException("A default outlet is allready registered!");
			else
				throw new ConfigurationException("An outlet with name " + outlet.getName() + " is allready registered!");
		}
		outlets.put(outlet.getName(), outlet);
	}

	protected FileHandle current() {
		return (fileHandles.isEmpty() ? null : fileHandles.peek());
	}
	
	/**
	 * DO NOT CALL THIS METHOD - FOR TESTS ONLY
	 */
	public FileHandle current__testONLY() {
		return current();
	}
	
	

	private boolean deleteLine = false;

	public void write(final String bytes) {
		if (current() != null) {
			if (deleteLine) {
				String temp = trimUntilNewline(bytes);
				removeWSAfterLastNewline(current().getBuffer());
				((StringBuffer) current().getBuffer()).append(temp);
			} else {
			    ((StringBuffer)current().getBuffer()).append(bytes);
			}
		}
		deleteLine = false;
	}

	public void removeWSAfterLastNewline(CharSequence cs) {
	    final StringBuffer buffer = (StringBuffer) cs;
		int i = buffer.length();
		boolean wsOnly = true;
		for (; i > 0 && wsOnly; i--) {
			char c = buffer.charAt(i - 1);
			wsOnly = Character.isWhitespace(c);
			if (wsOnly && isNewLine(c)) {
				buffer.delete(i, buffer.length());
				return;
			}
		}
		return;
	}

	private boolean isNewLine(char c) {
		return c == '\n' || c == '\r';
	}

	public String trimUntilNewline(String bytes) {
		int i = 0;
		boolean wsOnly = true;
		for (; i < bytes.length() && wsOnly; i++) {
			char c = bytes.charAt(i);
			wsOnly = Character.isWhitespace(c);
			if (wsOnly && isNewLine(c)) {
				if (c == '\r' && i + 1 < bytes.length() && bytes.charAt(i + 1) == '\n')
					i++;
				return bytes.substring(i + 1);
			}
		}
		return bytes;
	}

	private final static Pattern p = Pattern.compile("(.+)://(.+)");

	public static Pair<Outlet, String> resolveOutlet (Map<String, Outlet> allOutlets, String path, String outletName) {
        if (outletName == null) {
            final Matcher m = p.matcher(path);
            if (m.matches()) {
                outletName = m.group(1);
                path = m.group(2);
            }
        }
        final Outlet o = allOutlets.get(outletName);
        if (o == null) {
            if (outletName == null)
                throw new ConfigurationException ("No default outlet was configured!");
            else
                throw new ConfigurationException ("No outlet with the name " + outletName + " could be found!");
        }
        
        return new Pair<Outlet, String> (o, path);
	}
	
	public void openFile(String path, String outletName) {
	    final Pair<Outlet, String> raw = resolveOutlet(outlets, path, outletName);
	    
	    final Outlet actualOutlet = raw.getFirst();
	    final String actualPath = raw.getSecond();
	    
		fileHandles.push (actualOutlet.createFileHandle (actualPath));
	}

	public void closeFile() {
		final FileHandle fi = fileHandles.pop();
		fi.writeAndClose();
	}

	private final Stack<SyntaxElement> s = new Stack<SyntaxElement>();

	public void pushStatement(final SyntaxElement stmt, XpandExecutionContext ctx) {
		if (stmt instanceof TextStatement) {
			deleteLine = ((TextStatement) stmt).isDeleteLine();
			if (automaticHyphenation) {
				deleteLine = true;
			}
		}
		s.push(stmt);
	}

	public SyntaxElement popStatement() {
		final SyntaxElement se = s.pop();
		return se;
	}

}